#!/usr/local/ai-contrib/depot/python-1.4/bin/python

import os, string, regex, types, sys, regsub, fnmatch

Maker = None

CircularStringDependency = "CircularStringDependency"
CircularMakeDependency = "CircularMakeDependency"
CantCreate = "CantCreate"

class PyMake:
    def __init__(self, makefile = "PyMakefile",
		 verbose = 1, true_execute = 1, make_cmd = None):
	self.makefile = makefile
	self._verbose = verbose
	self.verbose = verbose or (not true_execute)
	if make_cmd:
	    if verbose > 1:
		make_cmd = make_cmd + " -vv"
	    elif not verbose: make_cmd = make_cmd + " -silent"
	    if not true_execute:
		make_cmd = make_cmd + " -n"
	self.true_execute = true_execute
	self.make_cmd = make_cmd
	self.expand_pat = regex.compile("%(\([A-Za-z0-9_]*\))")
    def Make(self, target = "all"):
	global Maker
	# First we load the rules and evaluate the context
	mf_context = {"MFRule": MFRule}
	self.rule_db = []
	Maker = self
	try:
	    execfile(self.makefile, mf_context)
	except IOError:
	    print "Error: encountered missing file while trying to load %s." % self.makefile
	    sys.exit()
	Maker = None
	self.mf_context = mf_context
	# Next we build the internal dictionary
	self.ComputeMFDictionary(mf_context)
	# Now, we look for a rule.
	top_rule, binding_dict = self.MatchMFRule(target)
	# If there's no rule, raise an error
	if not top_rule:
	    raise CantCreate, target
	# Now, use the rule to construct a plan
	top_rule.MaybeExecuteBody(binding_dict, {})
    def ComputeMFDictionary(self, mf_context):
	self.fleshing_dict = {}
	for key, value in mf_context.items():
	    # Copy over all strings.
	    if type(value) is types.StringType:
		self.fleshing_dict[key] = value
	    elif type(value) is types.ListType:
		# Copy over all lists of strings, as joined strings
		all_strings = 1
		for element in value:
		    if type(element) is not types.StringType:
			all_strings = 0
			break
		if all_strings:
		    self.fleshing_dict[key] = string.join(value)
	# Add things like MAKE.
	if self.make_cmd:
	    self.fleshing_dict["MAKE"] = self.make_cmd
	# Now that the dictionary is constructed, I expand all
	# the strings in advance, so I don't have to recurse later.
	for key, value in self.fleshing_dict.items():
	    self.fleshing_dict[key] = self.ExpandMFStringGlobal(value, {}, [])
    def MatchMFRule(self, target):	
	for element in self.rule_db:
	    if self.verbose > 1:
		if type(element.target) is types.StringType:
		    pat = element.target
		else:
		    pat = element.target.givenpat
		print "Target %s, rule target %s" % (target, pat)
	    binding_dict = element.Match(target)
	    if self.verbose > 1:
		if binding_dict:
		    print "...matched."
		else: print "...failed."
	    if binding_dict:
		return element, binding_dict
	return None, None
    def ExpandMFString(self, str):
	return self.ExpandMFStringGlobal(str, {}, [], 0)
    def ExpandMFStringGlobal(self, str, subst_dict, already_replaced, recursing = 1):
	# I must do a number of things here. In particular, because I don't
	# want people to have to type %(VAR)s all over the place, I reduce it
	# to %(VAR) in the input and modify here. This also allows me to
	# search for recursion.
	dict = self.fleshing_dict
	res = 0
	while 1:
	    # We search for the match pattern. 
	    res = self.expand_pat.search(str, res)
	    # If there are no more matches, end
	    if res < 0: break
	    match = self.expand_pat.group(0)
	    # We retrieve the element name.
	    element_name = self.expand_pat.group(1)
	    # If the element name has already been replaced, well,
	    # we're in trouble.
	    if element_name in already_replaced:
		raise CircularStringDependency, element_name
	    # Otherwise, if the element name is already
	    # in the subst dictionary, we don't need to expand it.
	    if not subst_dict.has_key(element_name):
		# Otherwise, look for the key in the fleshing dictionary.
		# If it's there, generate the replacement string for
		# the element and substitute in. Otherwise, add a %s
		# and move on.
		if dict.has_key(element_name):
		    if recursing:
			newstr = self.ExpandMFStringGlobal(dict[element_name],
							   subst_dict,
							   already_replaced + [element_name])
			dict[element_name] = newstr
		    else:
			newstr = dict[element_name]
		    subst_dict[element_name] = (match, newstr)
		else:
		    subst_dict[element_name] = (match, match + "s")
	    # Increment the search starting point and loop
	    res = res + len(match)
	# Now, make the substitutions and return the string.
	for old, new in subst_dict.values():
	    str = regsub.gsub(old, new, str)
	return str	    
	    
class MFRule:
    def __init__(self, target, dependencies, body, restriction_list = []):
	global Maker
	self.target = target
	self.dependencies = dependencies
	# Body is a string, a list of strings, or a function
	# or lambda.
	self.maker = Maker
	if type(body) is types.StringType:
	    self.body = PyMakeRuleStringBody(self, [body])
	elif type(body) is types.ListType:
	    self.body = PyMakeRuleStringBody(self, body)
	else: self.body = PyMakeRuleFnBody(self, body)
	self.expanded = 0
	# Insert the rule at the front of the list
	self.maker.rule_db[0:0] = [self]
	# The restriction list is a first-pass filter on eligible files.
	# So if you want to rules to make .o from .c, you can
	# filter them based on the restriction list.	
	self.restriction_list = restriction_list
    def ExpandFilename(self, fname):
	pat = self.maker.ExpandMFString(fname)
	return regex.symcomp(fnmatch.translate(pat))
    def Expand(self):	
	self.target = self.ExpandFilename(self.target)
	self.dependencies = map(self.maker.ExpandMFString, self.dependencies)
	self.body.ExpandBody()
	self.restriction_list = map(self.ExpandFilename, self.restriction_list)
	self.expanded = 1
    def Match(self, input_target):
	if not self.expanded:
	    self.Expand()
	# If there are resetrictions, the input_target must match one of them.
	if self.restriction_list:
	    reject = 1
	    for restr_pat in self.restriction_list:
		if self.maker.verbose > 1:
		    sys.stdout.write("Filtering by %s..." % restr_pat.givenpat)
		if restr_pat.match(input_target) == len(input_target):
		    reject = 0
		    break
	    if reject:
		return None
	if self.target.match(input_target) == len(input_target):
	    # If the entire input matches, then construct the
	    # binding dictionary.
	    root, ext = os.path.splitext(input_target)
	    if ext: ext = ext[1:]
	    dict = {"FILE": input_target, "BASE": root, "EXT": ext}
	    # Also get the symbolic matches from the compiled regex.
	    for key, value in self.target.groupindex.items():
		dict[key] = self.target.group(value)
	    return dict
	else: return None
    def ExpandLocalString(self, str, binding_dict):
	while 1:
	    try:
		return (str % binding_dict)
	    except KeyError, key:
		binding_dict[key] = ""
    def ExpandLocalBody(self, binding_dict):
	return map(lambda x, s = self, b = binding_dict: string.strip(s.ExpandLocalString(x, b)),
		   self.body)
    # A return value of 1 means that something needed to be done,
    # a return value of 0 means that nothing needed to be done
    # a return value of -1 means that something went wrong
    def MaybeExecuteBody(self, binding_dict, already_scheduled, plan_stack = []):
	target = binding_dict["FILE"]
	if already_scheduled.has_key(target):
	    return already_scheduled[target]
	elif target in plan_stack:
	    raise CircularMakeDependency, target
	if not self.dependencies:
	    # Absolute rule, always execute.
	    val = self.body.Execute(binding_dict)
	    already_scheduled[target] = val
	    return val
	else:
	    # There are dependencies.
	    execute_body = 0
	    # If the target doesn't exist, execute the body
	    if not os.path.exists(target):
		target_exists = 0
		execute_body = 1
	    else:
		target_exists = 1
		mtime_target = os.stat(target)[8]
	    for dependency in self.dependencies:
		# First, expand the dependency based on the match
		true_file = self.ExpandLocalString(dependency, binding_dict)
		# In the future, we might do glob matching against the
		# file system, but not right now.
		# So next step is to try to find a match for this new
		# target.
		dependency_rule, dep_binding_dict = self.maker.MatchMFRule(true_file)
		if dependency_rule:
		    # If a match is found, try to find a plan.
		    status = dependency_rule.MaybeExecuteBody(dep_binding_dict, already_scheduled,
							      plan_stack + [target])
		    if status == -1:
			# Something went wrong.
			already_scheduled[target] = -1
			return -1
		    elif status:
			# If one of the dependencies was built, then
			# we have something we must do,
			# and we choose to execute the body.
			# If the dependency didn't need anything done,
			# there's nothing to update.
			execute_body = 1
		else:
		    # If there's no rule, we check to see if the dependency
		    # exists.
		    if os.path.exists(true_file):
			if target_exists:
			    mtime_dep = os.stat(true_file)[8]
			    if mtime_dep > mtime_target:
				execute_body = 1
		    else:
			raise CantCreate, true_file
	    if execute_body:
		val = self.body.Execute(binding_dict)
		already_scheduled[target] = val
		return val
	    else:
		already_scheduled[target] = 0
		return 0

class PyMakeRuleBody:
    def __init__(self, rule):
	self.rule = rule
	self.maker = rule.maker
    def System(self, str):
	# This doesn't take into account error state or
	# customizing the shell.
	os.system(str)

class PyMakeRuleStringBody(PyMakeRuleBody):
    def __init__(self, rule, string_list):
	self.string_list = string_list
	PyMakeRuleBody.__init__(self, rule)
    def ExpandBody(self):
	self.string_list = map(self.maker.ExpandMFString, self.string_list)
    def Execute(self, binding_dict):
	for str in self.string_list:
	    newstr = string.strip(self.rule.ExpandLocalString(str, binding_dict))
	    if self.maker.verbose:
		print newstr
	    if self.maker.true_execute:
		# This doesn't control for which shell!
		# Nor does it terminate processing when an error is
		# encountered.
		self.System(newstr)
	return 1

PyMakeBodyError = "PyMakeBodyError"
	
class PyMakeRuleFnBody(PyMakeRuleBody):
    def __init__(self, rule, fn):
	self.fn = fn	
	self.binding_dict = {}
	PyMakeRuleBody.__init__(self, rule)
    def ExpandBody(self):
	pass
    def Execute(self, binding_dict):
	if self.maker.verbose:
	    print "Executing body of rule `%s' for target %s" % (self.rule.target.givenpat,
								 binding_dict["FILE"])
	self.binding_dict = binding_dict	
	try:
	    self.fn(self)
	    self.binding_dict = {}
	    return 1
	except PyMakeBodyError:
	    self.binding_dict = {}
	    return -1
	except PyMakeBodyError:
	    self.binding_dict = {}
	    print "Encountered error:", sys.exc_type, sys.exc_value	    
	    return -1
    # Methods below here are for contents of the Execute fn.
    def SystemCall(self, str):
	str = self.Expand(str)
	if self.maker.verbose:
	    print "System call: %s" % str
	if self.maker.true_execute:
	    self.System(str)
    def Execv(self, *argtuple):
	args = tuple(map(self.Expand, argtuple[1:]))
	if self.maker.verbose:
	    print "Call: %s(%s)" % (`argtuple[0]`, string.joinfields(map(repr, args), ", "))
	if self.maker.true_execute:	
	    apply(argtuple[0], args)
    def Lookup(self, str):
	return self.maker.mf_context[str]
    def Expand(self, str):
	if type(str) is types.StringType:
	    return self.rule.ExpandLocalString(self.maker.ExpandMFString(str), self.binding_dict)
	else: return str
    def InDirectory(self, dir, *tuples):
	if self.maker.verbose:
	    print "Entering directory: %s" % dir
	if self.maker.true_execute:
	    raise_p = 0
	    cwd = os.getcwd()
	    os.chdir(dir)
	    try:
		for tuple in tuples:
		    apply(self.Execv, tuple)
	    except PyMakeBodyError:
		os.chdir(cwd)
		raise_p = 1
	    except PyMakeBodyError:
		os.chdir(cwd)
		print "Encountered error:", sys.exc_type, sys.exc_value
		raise_p = 1
	    os.chdir(cwd)
	    if raise_p:
		raise PyMakeBodyError
	if self.maker.verbose:
	    print "Leaving directory: %s" % dir
    def Make(self, target):
	# Always call make
	if self.maker.verbose:
	    print "Making: %s" % target
	makeobj = PyMake(verbose = self.maker._verbose,
			 true_execute = self.maker.true_execute,
			 make_cmd = self.maker.make_cmd)	
	makeobj.Make(target)

def Usage():
    print "Usage: pymake.py [-f makefile] [-vv] [-silent] [-n] [target]"
    sys.exit()

if __name__ == '__main__':
    command = sys.argv[0]
    args = sys.argv[1:]
    verbose = 1
    true_execute = 1
    makefile = "PyMakefile"
    while args:		
	if args[0] == "-silent":
	    verbose = 0
	    args = args[1:]
	elif args[0] == "-vv":
	    verbose = 2
	    args = args[1:]
	elif args[0] == "-n":
	    true_execute = 0
	    args = args[1:]
	elif args[0] == "-f":
	    if len(args) == 1:
		Usage()
	    else:
		makefile = args[1]
		args = args[2:]
	else:
	    break
    if len(args) > 1:
	Usage()
    elif not args:
	target = "all"
    else:
	target = args[0]
    makeobj = PyMake(makefile, verbose, true_execute, command)
    makeobj.Make(target)
